package it.unibo.tw0708.dao.hsqldb;

import it.unibo.tw0708.dao.UserDAO;
import it.unibo.tw0708.dao.UserDTO;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;

/**
 * Implementazione di query jdbc su una versione hqldb della tabella user
 * 
 * La scrittura dei metodi DAO con JDBC segue sempre e comunque questo schema:
 * 
	// --- 1. Dichiarazione della variabile per il risultato ---
	// --- 2. Controlli preliminari sui dati in ingresso ---
	// --- 3. Apertura della connessione ---
	// --- 4. Tentativo di accesso al db e impostazione del risultato ---
	// --- 5. Gestione di eventuali eccezioni ---
	// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
	// --- 7. Restituzione del risultato (eventualmente di fallimento)
 * 
 * Inoltre, all'interno di uno statement JDBC si segue lo schema generale
 * 
	// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
	// --- b. Pulisci e imposta i parametri (se ve ne sono)
	// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
	// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
	// --- e. Rilascia la struttura dati del risultato
	// --- f. Rilascia la struttura dati dello statement
 * 
 * @author pisi79
 *
 */
public class HsqldbUserDAO 
implements UserDAO {

	// === Costanti letterali per non sbagliarsi a scrivere !!! ============================
	
	static final String user_table = "user";

	// -------------------------------------------------------------------------------------

	static final String user_id = "id";
	static final String user_name = "name";
	static final String user_surname = "surname";
	static final String user_address = "address";
	static final String user_matriculation = "matriculation";
	static final String user_nick = "nick";
	static final String user_password = "password";
	
	// == STATEMENT SQL ====================================================================

	// INSERT INTO usertable ( namecolumn,surnamecolumn, ...) VALUES ( ?,?, ... );
	static final String insert = 
		"INSERT " +
			"INTO " + user_table + " ( " + 
				user_name +", "+user_surname+", "+user_address+", "+user_matriculation+", "+user_nick+", "+user_password + " " +
			") " +
			"VALUES (?,?,?,?,?,?) ;"
		;
	
	// SELECT * FROM usertable WHERE idcolumn = ?;
	static String read_by_id = 
		"SELECT * " +
			"FROM " + user_table + " " +
			"WHERE " + user_id + " = ? ;"
		;

	// SELECT * FROM usertable WHERE nickcolumn = ?;
	static String read_by_nick = 
		"SELECT * " +
			"FROM " + user_table + " " +
			"WHERE " + user_nick + " = ? ;"
		;

	// DELETE FROM usertable WHERE idcolumn = ?;
	static String delete = 
		"DELETE " +
			"FROM " + user_table + " " +
			"WHERE " + user_id + " = ? ;"
		;

	// UPDATE usertable SET namecolumn = ?, surnamecolumn = ?, ... WHERE idcolumn = ?;
	static String update = 
		"UPDATE " + user_table + " " +
			"SET " + 
				user_name + " = ?, " +
				user_surname + " = ?, " +
				user_address + " = ?, " +
				user_matriculation + " = ?, " +
				user_nick + " = ?, " +
				user_password + " = ? " +
			"WHERE " + user_id + " = ? ;"
		;

	// SELECT * FROM usertable;
	static String query = 
		"SELECT * " +
			"FROM " + user_table + " ;"
		;

	// -------------------------------------------------------------------------------------

	// CALL IDENTITY();
	/**
	 * Restituisce l'identificativo dell'ultima tupla inserita. FUNZIONE NON CROSS-DATABASE!!!
	 */
	static String lastInsert = 
		"CALL " + 
			"IDENTITY() ;"
		;

	// -------------------------------------------------------------------------------------

	// CREATE usertable ( idcolumn INT NOT NULL IDENTITY, namecolumn VARCHAR(50), ..., UNIQUE ( idcolumn ) );
	static String create = 
		"CREATE " +
			"TABLE " + user_table +" ( " +
				user_id + " INT NOT NULL IDENTITY, " +
				user_name + " VARCHAR(50), " +
				user_surname + " VARCHAR(50), " +
				user_address + " LONGVARCHAR, " +
				user_matriculation + " BIGINT, " +
				user_nick + " VARCHAR(25), " +
				user_password + " BINARY, " +
				"UNIQUE( " + user_id + " ) " +
			") ;"
		;
	static String drop = 
		"DROP " +
			"TABLE " + user_table + " ;"
		;
	
	// === METODI DAO =========================================================================
	
	/**
	 * C
	 */
	public int createUser(UserDTO t) {
		// --- 1. Dichiarazione della variabile per il risultato ---
		int result = -1;
		// --- 2. Controlli preliminari sui dati in ingresso ---
		if ( t == null )  {
			System.err.println(HsqldbUserDAO.class.getName()+".createUser(): failed to insert a null user");
			return result;
		}
		// --- 3. Apertura della connessione ---
		Connection conn = HsqldbDAOFactory.createConnection();
		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			PreparedStatement prep_stmt = conn.prepareStatement(HsqldbUserDAO.insert);
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			prep_stmt.clearParameters();
			prep_stmt.setString(1, t.getName());
			prep_stmt.setString(2, t.getSurname());
			prep_stmt.setString(3, t.getAddress());
			prep_stmt.setLong(4, t.getMatriculation());
			prep_stmt.setString(5, t.getNick());
			prep_stmt.setBytes(6, t.getPassword());
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			prep_stmt.executeUpdate();
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			// n.d.
			// --- e. Rilascia la struttura dati del risultato
			// n.d.
			// --- f. Rilascia la struttura dati dello statement
			prep_stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			System.err.println(HsqldbUserDAO.class.getName()+".createUser(): failed to insert user\n"+e);
			e.printStackTrace();
			result = -2;
		}

		// Nel caso della creazione di una nuova tupla eseguo un secondo accesso per sapere che id le è stato assegnato
		// Chiaramente è inutile farlo se già il primo accesso ha prodotto errori
		// --- 6./7. Chiusura della connessione in caso di errori e restituizione prematura di un risultato di fallimento
		if ( result == -2 ) {
			HsqldbDAOFactory.closeConnection(conn);
			return result;
		}
		
		// --- 1. Dichiarazione della variabile per il risultato ---
		// riutilizziamo quella di prima

		// --- 2. Controlli preliminari sui dati in ingresso ---
		// già fatti

		// --- 3. Apertura della connessione ---
		// ce n'è una già aperta, se siamo qui

		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			Statement stmt = conn.createStatement();
			ResultSet rs = stmt.executeQuery(HsqldbUserDAO.lastInsert);
			if ( rs.next() ) {
				result = rs.getInt(1);
			}		
			rs.close();
			stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			System.err.println(HsqldbUserDAO.class.getName()+".createUser(): failed to retrieve id of last inserted user\n"+e);
			e.printStackTrace();
		}
		// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
		finally {
			HsqldbDAOFactory.closeConnection(conn);
		}
		// --- 7. Restituzione del risultato (eventualmente di fallimento)
		return result;
	}

	// -------------------------------------------------------------------------------------

	/**
	 * R
	 */
	public UserDTO readUser(int id) {
		// --- 1. Dichiarazione della variabile per il risultato ---
		UserDTO result = null;
		// --- 2. Controlli preliminari sui dati in ingresso ---
		if ( id < 0 )  {
			System.err.println(HsqldbUserDAO.class.getName()+".readUser(): failed to read a user with negative id");
			return result;
		}
		// --- 3. Apertura della connessione ---
		Connection conn = HsqldbDAOFactory.createConnection();
		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			PreparedStatement prep_stmt = conn.prepareStatement(HsqldbUserDAO.read_by_id);
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			prep_stmt.clearParameters();
			prep_stmt.setInt(1, id);
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			ResultSet rs = prep_stmt.executeQuery();
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			if ( rs.next() ) {
				UserDTO user = new UserDTO();
				user.setId(rs.getInt(user_id));
				user.setName(rs.getString(user_name));
				user.setSurname(rs.getString(user_surname));
				user.setAddress(rs.getString(user_address));
				user.setMatriculation(rs.getLong(user_matriculation));
				user.setNick(rs.getString(user_nick));
				user.setPassword(rs.getBytes(user_password));
				result = user;
			}
			// --- e. Rilascia la struttura dati del risultato
			rs.close();
			// --- f. Rilascia la struttura dati dello statement
			prep_stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			System.err.println(HsqldbUserDAO.class+".readUser(): failed to retrieve user with id = " + id+"\n"+e);
			e.printStackTrace();
		}
		// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
		finally {
			HsqldbDAOFactory.closeConnection(conn);
		}
		// --- 7. Restituzione del risultato (eventualmente di fallimento)
		return result;
	}

	// -------------------------------------------------------------------------------------

	/**
	 * U
	 */
	public boolean updateUser(UserDTO t) {
		// --- 1. Dichiarazione della variabile per il risultato ---
		boolean result = false;
		// --- 2. Controlli preliminari sui dati in ingresso ---
		if ( t == null )  {
			System.out.println(HsqldbUserDAO.class.getName()+".updateUser(): failed to update a null user");
			return result;
		}
		// --- 3. Apertura della connessione ---
		Connection conn = HsqldbDAOFactory.createConnection();
		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			PreparedStatement prep_stmt = conn.prepareStatement(HsqldbUserDAO.update);
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			prep_stmt.clearParameters();
			prep_stmt.setString(1, t.getName());
			prep_stmt.setString(2, t.getSurname());
			prep_stmt.setString(3, t.getAddress());
			prep_stmt.setLong(4, t.getMatriculation());
			prep_stmt.setString(5, t.getNick());
			prep_stmt.setBytes(6, t.getPassword());
			prep_stmt.setInt(7, t.getId());
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			prep_stmt.executeUpdate();
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			// n.d. qui devo solo dire al chiamante che è andato tutto liscio
			result = true;
			// --- e. Rilascia la struttura dati del risultato
			// n.d.
			// --- f. Rilascia la struttura dati dello statement
			prep_stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			System.err.println(HsqldbUserDAO.class.getName()+".insertUser(): failed to update user\n"+e);
			e.printStackTrace();
		}
		// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
		finally {
			HsqldbDAOFactory.closeConnection(conn);
		}
		// --- 7. Restituzione del risultato (eventualmente di fallimento)
		return result;
	}

	// -------------------------------------------------------------------------------------

	/**
	 * D
	 */
	public boolean deleteUser(UserDTO t) {
		// --- 1. Dichiarazione della variabile per il risultato ---
		boolean result = false;
		// --- 2. Controlli preliminari sui dati in ingresso ---
		if ( t == null )  {
			System.err.println(HsqldbUserDAO.class.getName()+".deleteUser(): failed to delete a null user");
			return result;
		}
		// --- 3. Apertura della connessione ---
		Connection conn = HsqldbDAOFactory.createConnection();
		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			PreparedStatement prep_stmt = conn.prepareStatement(HsqldbUserDAO.delete);
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			prep_stmt.clearParameters();
			prep_stmt.setInt(1, t.getId());
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			prep_stmt.executeUpdate();
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			// n.d. Qui devo solo dire al chiamante che è andato tutto liscio
			result = true;
			// --- e. Rilascia la struttura dati del risultato
			// n.d.
			// --- f. Rilascia la struttura dati dello statement
			prep_stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			System.err.println(HsqldbUserDAO.class.getName()+".deleteUser(): failed to delete user with id = " + t.getId()+"\n"+e);
			e.printStackTrace();
		}
		// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
		finally {
			HsqldbDAOFactory.closeConnection(conn);
		}
		// --- 7. Restituzione del risultato (eventualmente di fallimento)
		return result;
	}

	// -------------------------------------------------------------------------------------

	/**
	 * Find by nick
	 */
	public UserDTO findUser(String nick) {
		// --- 1. Dichiarazione della variabile per il risultato ---
		UserDTO result = null;
		// --- 2. Controlli preliminari sui dati in ingresso ---
		if ( nick == null || nick.equals("") ) {
			System.err.println(HsqldbUserDAO.class.getName()+".deleteUser(): failed to read a user with no nick");
			return result;
		}
		// --- 3. Apertura della connessione ---
		Connection conn = HsqldbDAOFactory.createConnection();
		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			PreparedStatement prep_stmt = conn.prepareStatement(HsqldbUserDAO.read_by_nick);
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			prep_stmt.clearParameters();
			prep_stmt.setString(1, nick);
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			ResultSet rs = prep_stmt.executeQuery();
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			if ( rs.next() ) {
				UserDTO user = new UserDTO();
				user.setId(rs.getInt(user_id));
				user.setName(rs.getString(user_name));
				user.setSurname(rs.getString(user_surname));
				user.setAddress(rs.getString(user_address));
				user.setMatriculation(rs.getLong(user_matriculation));
				user.setNick(rs.getString(user_nick));
				user.setPassword(rs.getBytes(user_password));
				result = user;
			}
			// --- e. Rilascia la struttura dati del risultato
			rs.close();
			// --- f. Rilascia la struttura dati dello statement
			prep_stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			System.err.println(HsqldbUserDAO.class+".findUser(): failed to retrieve user with nick = " + nick+"\n"+e);
			e.printStackTrace();
		}
		// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
		finally {
			HsqldbDAOFactory.closeConnection(conn);
		}
		// --- 7. Restituzione del risultato (eventualmente di fallimento)
		return result;
	}

	// -------------------------------------------------------------------------------------

	public Collection<UserDTO> getUsers() {
		// --- 1. Dichiarazione della variabile per il risultato ---
		Collection<UserDTO> result = new ArrayList<UserDTO>();
		// --- 2. Controlli preliminari sui dati in ingresso ---
		// n.d.
		// --- 3. Apertura della connessione ---
		Connection conn = HsqldbDAOFactory.createConnection();
		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			Statement stmt = conn.createStatement();
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			// n.d.
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			ResultSet rs = stmt.executeQuery(query);
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			while (rs.next()) {
				UserDTO user = new UserDTO();
				user.setId(rs.getInt(user_id));
				user.setName(rs.getString(user_name));
				user.setSurname(rs.getString(user_surname));
				user.setAddress(rs.getString(user_address));
				user.setMatriculation(rs.getLong(user_matriculation));
				user.setNick(rs.getString(user_nick));
				user.setPassword(rs.getBytes(user_password));
				result.add(user);
			}
			// --- e. Rilascia la struttura dati del risultato
			rs.close();
			// --- f. Rilascia la struttura dati dello statement
			stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			System.err.println(HsqldbUserDAO.class.getName()+".getUsers(): failed to get users\n"+e);
			e.printStackTrace();
		}
		// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
		finally {
			HsqldbDAOFactory.closeConnection(conn);
		}
		// --- 7. Restituzione del risultato (eventualmente di fallimento)
		return result;
	}

	// -------------------------------------------------------------------------------------
	// -------------------------------------------------------------------------------------

	/**
	 * Creazione della table
	 */
	public boolean createTable() {
		// --- 1. Dichiarazione della variabile per il risultato ---
		boolean result = false;
		// --- 2. Controlli preliminari sui dati in ingresso ---
		// n.d.
		// --- 3. Apertura della connessione ---
		Connection conn = HsqldbDAOFactory.createConnection();
		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			Statement stmt = conn.createStatement();
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			// n.d.
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			stmt.execute(create);
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			// n.d. Qui devo solo dire al chiamante che è andato tutto liscio
			result = true;
			// --- e. Rilascia la struttura dati del risultato
			// n.d.
			// --- f. Rilascia la struttura dati dello statement
			stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			System.err.println(HsqldbUserDAO.class.getName()+".createTable(): failed to create user table\n"+e);
		}
		// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
		finally {
			HsqldbDAOFactory.closeConnection(conn);
		}
		// --- 7. Restituzione del risultato (eventualmente di fallimento)
		return result;
	}

	// -------------------------------------------------------------------------------------

	/**
	 * Rimozione della table
	 */
	public boolean dropTable() {
		// --- 1. Dichiarazione della variabile per il risultato ---
		boolean result = false;
		// --- 2. Controlli preliminari sui dati in ingresso ---
		// n.d.
		// --- 3. Apertura della connessione ---
		Connection conn = HsqldbDAOFactory.createConnection();
		// --- 4. Tentativo di accesso al db e impostazione del risultato ---
		try {
			// --- a. Crea (se senza parametri) o prepara (se con parametri) lo statement
			Statement stmt = conn.createStatement();
			// --- b. Pulisci e imposta i parametri (se ve ne sono)
			// n.d.
			// --- c. Esegui l'azione sul database ed estrai il risultato (se atteso)
			stmt.execute(drop);
			// --- d. Cicla sul risultato (se presente) pe accedere ai valori di ogni sua tupla
			// n.d. Qui devo solo dire al chiamante che è andato tutto a posto.
			result = true;
			// --- e. Rilascia la struttura dati del risultato
			// n.d.
			// --- f. Rilascia la struttura dati dello statement
			stmt.close();
		}
		// --- 5. Gestione di eventuali eccezioni ---
		catch (Exception e) {
			System.err.println(HsqldbUserDAO.class.getName()+".dropTable(): failed to drop user table\n"+e);
		}
		// --- 6. Rilascio, SEMPRE E COMUNQUE, la connessione prima di restituire il controllo al chiamante
		finally {
			HsqldbDAOFactory.closeConnection(conn);
		}
		// --- 7. Restituzione del risultato (eventualmente di fallimento)
		return result;
	}

}